import yaml
import os
import logging
import json


_PIPELINE_ROOT = "./config/pipeline"
_EXT = "yaml"
_ENVIRONMENT = "YETL_ENVIRONMENT"
_VAR_CWD = "{{cwd}}"


def _mkpath(name: str):
    """Function that builds the pipeline path"""

    env = os.getenv(_ENVIRONMENT)
    if not env:
        env = "local"

    path = f"{_PIPELINE_ROOT}/{env}/{name}.{_EXT}"
    path = os.path.abspath(path)
    return path


def load_pipeline_config(app_name: str, pipeline_name: str):
    """Loads a spark configuration to load data from landing to raw."""
    _logger = logging.getLogger(app_name)
    path = _mkpath(pipeline_name)
    _logger.info(f"Loading Dataflow configuration from file {path}")
    with open(path, "r") as f:
        pipeline = yaml.safe_load(f.read())

    _logger.debug(json.dumps(pipeline, indent=4, sort_keys=True, default=str))

    return pipeline


def load_config(app_name: str):
    _logger = logging.getLogger(app_name)
    path = _mkpath("config")
    _logger.info(f"Loading Dataflow configuration from file {path}")
    with open(path, "r") as f:
        config = f.read()

    config = config.replace(_VAR_CWD, os.getcwd())
    config = yaml.safe_load(config)
    _logger.debug(json.dumps(config, indent=4, sort_keys=True, default=str))

    return config



import json
import yaml
from ._ischema_repo import ISchemaRepo
from pyspark.sql.types import StructType
from ..file_system import FileFormat, IFileSystem, file_system_factory, FileSystemType
import os


class SchemaNotFound(Exception):
    def __init__(self, path: str):
        self.path = path
        super().__init__(self.path)


class SparkFileSchemaRepo(ISchemaRepo):

    _SCHEMA_ROOT = "./config/schema/spark"
    _EXT = "yaml"

    def __init__(self, context, config: dict) -> None:
        super().__init__(context, config)
        self.root_path = config["spark_schema_file"].get("spark_schema_root")

    def _mkpath(self, database_name: str, table_name: str):
        """Function that builds the schema path"""
        if not self.root_path:
            return f"{self._SCHEMA_ROOT}/{database_name}/{table_name}.{self._EXT}"
        else:
            return f"{self.root_path}/{database_name}/{table_name}.{self._EXT}"

    def save_schema(self, schema: StructType, database_name: str, table_name: str):
        """Serialise a spark schema to a yaml file and saves to a schema file in the schema folder."""

        path = self._mkpath(database_name, table_name)
        path = os.path.abspath(path)

        schema_dict = json.loads(schema.json())
        with open(path, "w") as f:
            f.write(yaml.safe_dump(schema_dict))

    def load_schema(self, database_name: str, table_name: str):
        """Loads a spark from a yaml file and deserialises to a spark schema."""

        path = self._mkpath(database_name, table_name)
        path = os.path.abspath(path)

        # this was in thought that schema's could be maintain and loaded off DBFS for databricks
        # it actually works much better using the local repo files
        # maybe considered for a future use case - we need more configuration to set the schema store
        # type, since it's explicitly hand in hand with databricks spark env.
        # fs: IFileSystem = self.context.fs

        # file system is working fine for databricks and vanilla spark deployments.
        fs: IFileSystem = file_system_factory.get_file_system_type(
            self.context, FileSystemType.FILE
        )

        self.context.log.info(
            f"Loading schema for dataset {database_name}.{table_name} from {path} using {type(fs)}"
        )
        try:
            schema = fs.read_file(path, FileFormat.YAML)
        except Exception as e:
            msg = f"Failed to load schema for dataset {database_name}.{table_name} from {path}"
            raise SchemaNotFound(path) from e

        msg = json.dumps(schema, indent=4, default=str)
        self.context.log.debug(msg)

        try:
            spark_schema = StructType.fromJson(schema)
        except Exception as e:
            msg = (
                msg
            ) = f"Failed to deserialise spark schema to StructType for dataset {database_name}.{table_name} from {path}"
            raise Exception(msg) from e

        return spark_schema



from ._destination import Destination


class StreamWriter(Destination):
    def __init__(
        self, context, database: str, table: str, dataset: dict, io_type: str
    ) -> None:
        super().__init__(context, database, table, dataset, io_type)



from pyspark.sql.types import StructType
from ..parser._constants import *
from . import _builtin_functions as builtin_funcs
from ..schema_repo import ISchemaRepo
from ..parser import parser
import uuid


class Dataset:
    def __init__(
        self, context, database: str, table: str, dataset: dict, io_type: str
    ) -> None:

        self.id = uuid.uuid4()
        self.datalake = dataset["datalake"]
        self.datalake_protocol = context.fs.datalake_protocol
        self.context = context
        self.database = database
        self.table = table
        self.database_table = f"{self.database}.{self.table}"
        self.path_date_format = dataset.get("path_date_format")
        self.file_date_format = dataset.get("file_date_format")
        self._path = self._get_path(dataset)
        self._timeslice_position = parser.get_slice_position(self.path, self)
        self._path = builtin_funcs.execute_replacements(self._path, self)
        self.context_id = dataset.get("context_id")
        self.timeslice = dataset.get("timeslice")

        # default format to delta if not
        fmt = dataset.get(FORMAT, Format.DELTA.name)
        self.format_type = Format[fmt.upper()]
        self._initial_load = False

    def _get_path(self, dataset: dict):
        path = dataset.get(PATH)
        return f"{self.datalake_protocol}{self.datalake}/{path}"

    @property
    def path(self):

        return self._path

    @property
    def initial_load(self):

        return self._initial_load

    @initial_load.setter
    def initial_load(self, value: bool):
        self._initial_load = value

    @property
    def format(self):
        return self.format_type.value

    def is_source(self):
        pass

    def is_destination(self):
        pass

    def save_metadata(self):
        self.context.metadata_repo.save(self)



from ._spark_file_schema_repo import SparkFileSchemaRepo, SchemaNotFound
from ._deltalake_sql_file import DeltalakeSchemaFile
from ._ischema_repo import ISchemaRepo
from ._factory import factory as schema_repo_factory

__all__ = [
    "ISchemaRepo",
    "SparkFileSchemaRepo",
    "DeltalakeSchemaFile",
    "schema_repo_factory",
    "SchemaNotFound",
]



from ._destination import Destination
from ..parser._constants import *
from ..schema_repo import ISchemaRepo
import os
from .. import _delta_lake as dl
from pyspark.sql import DataFrame
from typing import ChainMap
from ..parser import parser


class Writer(Destination):
    def __init__(
        self, context, database: str, table: str, config: dict, io_type: str
    ) -> None:
        super().__init__(context, database, table, config, io_type)

        self.dataframe: DataFrame = None
        # try and load a schema if schema on read
        self.table_ddl: str = self._get_table_sql(config)
        self.zorder_by: list = self._get_table_zorder(config)
        self.context.log.debug(f"Writer table ddl = {self.table_ddl}")

        # get the configured partitions.
        self.partitions: list = self._get_conf_partitions(config, self.table_ddl)
        # get the configured table property for auto optimise.
        self.auto_optimize = self._get_conf_dl_property(
            config, dl.DeltaLakeProperties.OPTIMIZE_WRITE
        )
        # get the configured table property for autocompact - delta.autoOptimize.autoCompact
        self.auto_compact = self._get_conf_dl_property(
            config, dl.DeltaLakeProperties.AUTO_COMPACT
        )

        # gets the read, write, etc options based on the type
        io_properties = config.get(io_type)
        self.auto_io = io_properties.get(AUTO_IO, True)
        self.options: dict = io_properties.get(OPTIONS)
        try:
            self.mode = io_properties[MODE]
        except KeyError as e:
            msg = f"{MODE} is missing from the Destination Write configuration and is required."
            context.log.error(msg)
            raise Exception(msg) from e

        self._initial_load = super().initial_load

        if self.auto_io:
            self.context.log.info(
                f"auto_io = {self.auto_io} automatically creating or altering delta table {self.database}.{self.table}"
            )

            # returns the table properties and partitions if the table already exists.
            current_properties = self.create_or_alter_table()

            # alter, drop or create any constraints defined that are not on the table
            self._set_table_constraints(current_properties, config)
            # alter, drop or create any properties that are not on the table
            self._set_table_properties(current_properties, config)

            # TODO: if the table partitions in the properties form the table is different to the partitions
            # in the config or ddl, yetl.allowRepartitioning = true then repartition the table
            # Is this a feature we want? Not sure.
            self._table_repartition(current_properties, config)

    def _table_repartition(self, table_properties: dict, config: dict):
        pass

    def _get_conf_partitions(self, config: dict, table_ddl: str):

        # get the partitioned columns from the SQL ddl
        partitions = None
        if table_ddl:
            try:
                partitions = parser.sql_partitioned_by(table_ddl)
                msg = f"Parsed partitioning columns from sql ddl for {self.database}.{self.table} as {partitions}"
                self.context.log.info(msg)
            except Exception as e:
                msg = f"An error has occured parsing sql ddl partitioned clause for {self.database}.{self.table} for the ddl: {table_ddl}"
                self.context.log.error(msg)
                raise Exception(msg) from e

        # if there is no sql ddl then take it from the yaml parameters
        if not partitions:
            table: dict = config[TABLE]
            # otherwise there are no partitioned columns and default to None
            partitions = table.get("partitioned_by", [])
            msg = f"Parsed partitioning columns from dataflow yaml config for {self.database}.{self.table} as {partitions}"

        return partitions

    def _set_table_constraints(self, table_properties: dict, config: dict):
        _existing_constraints = {}
        if table_properties:
            _existing_constraints = table_properties.get(self.database_table)
            _existing_constraints = _existing_constraints.get("constraints")

        self.column_constraints_ddl = self._get_check_constraints_sql(
            config, _existing_constraints
        )
        self.context.log.debug(
            f"Writer table check constraints ddl = {self.column_constraints_ddl}"
        )
        if self.table_ddl or not self.initial_load:
            # can only add constraints to columns if there are any
            # if there is no table_ddl an empty table is created and the data schema defines the table
            # on the initial load so this is skipped on the 1st load.
            if self.column_constraints_ddl:
                for cc in self.column_constraints_ddl:
                    self.context.spark.sql(cc)

    def _set_table_properties(self, existing_properties: dict, config: dict):
        _existing_properties = {}
        if existing_properties:
            _existing_properties = existing_properties.get(self.database_table)
            _existing_properties = _existing_properties.get(PROPERTIES)
        self.tbl_properties_ddl = self._get_table_properties_sql(
            config, _existing_properties
        )
        self.context.log.debug(
            f"Writer table properties ddl = {self.tbl_properties_ddl}"
        )
        if self.tbl_properties_ddl:
            self.context.spark.sql(self.tbl_properties_ddl)

    def create_or_alter_table(self):

        properties = None
        dl.create_database(self.context, self.database)
        table_exists = dl.table_exists(self.context, self.database, self.table)
        if table_exists:
            self.context.log.info(
                f"Table already exists {self.database}.{self.table} at {self.path}"
            )
            self.initial_load = False
            # on non initial loads get the constraints and properties
            # to them to and sync with the declared constraints and properties.
            # TODO: consolidate details and properties fetch since the properties are in the details. The delta lake api may have some improvements.
            properties = dl.get_table_properties(
                self.context, self.database, self.table
            )
            details = dl.get_table_details(self.context, self.database, self.table)
            # get the partitions from the table details and add them to the properties.
            table_name = f"{self.database}.{self.table}"
            properties[table_name][PARTITIONS] = details[table_name][PARTITIONS]

        else:
            dl.create_table(
                self.context, self.database, self.table, self.path, self.table_ddl
            )
            self.initial_load = True

        return properties

    def _get_conf_dl_property(self, config: dict, property: dl.DeltaLakeProperties):
        return (
            config.get(TABLE)
            and config.get(TABLE).get(PROPERTIES)
            and config.get(TABLE).get(PROPERTIES).get(property.value)
        )

    def _get_check_constraints_sql(
        self, config: dict, existing_constraints: dict = None
    ):
        table = config.get(TABLE)
        sql_constraints = []
        if table:
            check_constraints = table.get(CHECK_CONSTRAINTS)

        if not check_constraints:
            check_constraints = {}

        # if the existing constraint is not defined in the config constraints
        # and it is different then drop it and recreate
        if existing_constraints:
            for name, existing_constraint in existing_constraints.items():
                defined_constraint = check_constraints.get(name)

                # if the existing constraint is not defined in the config constraints then drop it
                if not defined_constraint:
                    sql_constraints.append(
                        dl.alter_table_drop_constraint(self.database, self.table, name)
                    )
                # if the existing constraint is defined and it is different then drop and add it
                elif (
                    defined_constraint.replace(" ", "").lower()
                    != existing_constraint.replace(" ", "").lower()
                ):
                    sql_constraints.append(
                        dl.alter_table_drop_constraint(self.database, self.table, name)
                    )
                    sql_constraints.append(
                        dl.alter_table_add_constraint(
                            self.database, self.table, name, defined_constraint
                        )
                    )

        # the constraint is defined but doesn't exist on the table yet so
        # add the constraint
        if check_constraints:
            for name, defined_constraint in check_constraints.items():
                existing_constraint = existing_constraints.get(name)
                if not existing_constraint:
                    sql_constraints.append(
                        dl.alter_table_add_constraint(
                            self.database, self.table, name, defined_constraint
                        )
                    )

        return sql_constraints

    def _get_table_properties_sql(self, config: dict, existing_properties: dict = None):

        table = config.get("table")
        if table:
            tbl_properties = table.get(PROPERTIES)
            if existing_properties:
                tbl_properties = dict(ChainMap(tbl_properties, existing_properties))
            tbl_properties = [f"'{k}' = '{v}'" for k, v in tbl_properties.items()]
            tbl_properties = ", ".join(tbl_properties)
            tbl_properties = dl.alter_table_set_tblproperties(
                self.database, self.table, tbl_properties
            )
            return tbl_properties
        else:
            return None

    def _get_table_zorder(self, config: dict):

        table = config.get(TABLE)
        zorder_by: list = []
        if table:
            zorder_by = table.get("zorder_by", [])

        return zorder_by

    def _get_table_sql(self, config: dict):

        table = config.get(TABLE)
        if table:
            ddl: str = table.get("ddl")
            if ddl and os.path.exists(ddl):
                self.schema_repo: ISchemaRepo = (
                    self.context.schema_repo_factory.get_schema_repo_type(
                        self.context, config["deltalake_schema_repo"]
                    )
                )
                ddl = self.schema_repo.load_schema(self.database, self.table)
                ddl = ddl.replace("{{database_name}}", self.database)
                ddl = ddl.replace("{{table_name}}", self.table)
                ddl = ddl.replace("{{path}}", self.path)
        else:
            ddl = None

        return ddl

    def write(self):
        self.context.log.info(f"Writing data to {self.database_table} at {self.path}")
        if self.dataframe:

            auto_compact = all(
                [self.auto_compact, self.partitions, not self.context.is_databricks]
            )
            if auto_compact:
                self.context.log.info(
                    f"Auto compacting in memory partitions for {self.database_table} on partitions {self.partitions}"
                )
                self.dataframe = self.dataframe.coalesce(1).repartition(
                    *self.partitions
                )

            (super().write())

            auto_optimize = all([self.auto_optimize, not self.context.is_databricks])
            if auto_optimize:
                self.context.log.info(
                    f"Auto optimizing {self.database_table} where {self.partition_values} zorder by {self.zorder_by}"
                )
                dl.optimize(
                    self.context,
                    self.database,
                    self.table,
                    self.partition_values,
                    self.zorder_by,
                )

        else:
            msg = f"Writer dataframe isn't set and cannot be written for {self.database_table} at {self.path}"
            self.context.log.error(msg)
            raise Exception(msg)

    @property
    def initial_load(self):

        return self._initial_load

    @initial_load.setter
    def initial_load(self, value: bool):

        if not self.options[MERGE_SCHEMA] and value:
            self.options[MERGE_SCHEMA] = value
        self._initial_load = value



from ..dataset import Dataset, Save, DefaultSave
from typing import Type
from abc import ABC, abstractmethod
from pyspark.sql import DataFrame
import json


class IDataflow(ABC):
    def __init__(
        self,
        context,
        config: dict,
        dataflow_config: dict,
        save_type: Type[Save] = DefaultSave,
    ) -> None:

        self.log = context.log
        self.log.debug("initialise dataflow with config")
        self.log.debug(json.dumps(config, indent=4, default=str))

        self.log.debug("initialise dataflow with dataflow_config")
        self.log.debug(json.dumps(dataflow_config, indent=4, default=str))

        self.datalake = config["datalake"]

        self._spark_schema_repo = config["spark_schema_repo"]
        self._deltalake_schema_repo = config["deltalake_schema_repo"]

        self.datalake_protocol = context.fs.datalake_protocol
        self.context = context
        self.sources = {}
        self.destinations = {}

        self.retries = dataflow_config.get("retries", 0)
        try:
            del dataflow_config["retries"]
        except KeyError:
            pass

        self.retry_wait = dataflow_config.get("retry_wait", 0)
        try:
            del dataflow_config["retry_wait"]
        except KeyError:
            pass

        self.enabled_dataflow_types = dataflow_config.get("enable_dataflow_types", [])
        try:
            del dataflow_config["enable_dataflow_types"]
        except KeyError:
            pass

    @abstractmethod
    def append(self, dataset: Dataset) -> None:
        pass

    @abstractmethod
    def source_df(self, database_table: str) -> DataFrame:
        pass

    @abstractmethod
    def destination_df(self, database_table: str, dataframe: DataFrame):
        pass



from sys import excepthook
from yetl_flow import yetl_flow, IDataflow, Reader, Writer, Context
from pyspark.sql.functions import *
import logging
from datetime import datetime


@yetl_flow(log_level="ERROR", auto_read=True)
def customer_landing_to_rawdb_csv(
    context: Context, dataflow: IDataflow, timeslice: datetime, timeslice_mask: str
) -> dict:
    """Load the demo customer data as is into a raw delta hive registered table.

    this is a test pipeline that can be run just to check everything is setup and configured
    correctly.
    """
    context.log.info(
        f"Executing Dataflow {context.app_name} with timeslice={timeslice}, retries={dataflow.retries}"
    )

    # the config for this dataflow has 2 landing sources that are joined
    # and written to delta table
    # delta tables are automatically created and if configured schema exceptions
    # are loaded syphened into a schema exception table
    df_cust = dataflow.source_df("landing.customer")
    df_prefs = dataflow.source_df("landing.customer_preferences")

    dst: Writer = dataflow.destinations["raw.customer"]

    # df = df_cust.join(df_prefs, "id", "inner")
    df = df_cust

    dataflow.log.info(f"Writing data to {dst.path}")
    (df.write.format(dst.format).options(**dst.options).mode(dst.mode).save(dst.path))


def test_integration():
    """Bare bones integration test until the project refactoring settles down"""
    logger = logging.getLogger(__name__)
    logger.setLevel(logging.INFO)
    logger.info(f"Executing pipeline test_customer_landing_to_rawdb")

    try:
        timeslice = datetime(2022, 1, 1, 0, 0, 0, 0)
        timeslice_mask = "*/*/*"
        results = customer_landing_to_rawdb_csv(
            timeslice=timeslice, timeslice_mask=timeslice_mask
        )
        # results = test_pipeline.test_customer_landing_to_rawdb_csv()
        assert True
    except Exception as e:
        raise



from pyspark.sql.dataframe import DataFrame
from pyspark.sql import functions as fn
from pyspark.sql import SparkSession
from pyspark.sql.utils import AnalysisException
from typing import Callable
from ..parser._constants import *


class IValidator:
    def __init__(
        self,
        context,
        dataframe: DataFrame,
        exceptions_handler: Callable[[DataFrame], int],
        database: str,
        table: str,
    ) -> None:
        self.context = context
        self.dataframe = dataframe
        self.exceptions_handler = exceptions_handler
        self.exceptions = None
        self.exceptions_count = 0
        self.valid_count = 0
        self.total_count = 0
        self.database = database
        self.table = table

    def validate(self) -> dict:
        pass

    def get_result(self):
        validation = {
            "validation": {
                "schema_on_read": {
                    f"{self.database}.{self.table}": {
                        "total_count": self.total_count,
                        "valid_count": self.valid_count,
                        "exception_count": self.exceptions_count,
                    }
                }
            }
        }
        return validation


class PermissiveSchemaOnRead(IValidator):
    def __init__(
        self,
        context,
        dataframe: DataFrame,
        exceptions_handler: Callable[[DataFrame], int],
        database: str,
        table: str,
    ) -> None:
        super().__init__(context, dataframe, exceptions_handler, database, table)
        self.database = database
        self.table = table

    def validate(self):

        self.total_count = self.dataframe.count()
        self.dataframe.cache()

        self.exceptions = (
            self.dataframe.where(f"{CORRUPT_RECORD} IS NOT NULL")
            .withColumn(TIMESTAMP, fn.current_timestamp())
            .withColumn(DATABASE, fn.lit(self.database))
            .withColumn(TABLE, fn.lit(self.table))
        )

        self.exceptions_count = self.exceptions.count()
        self.dataframe = self.dataframe.where(f"{CORRUPT_RECORD} IS NULL").drop(
            CORRUPT_RECORD
        )

        self.valid_count = self.dataframe.count()
        self.exceptions_count = self.exceptions_handler(self.exceptions)

        return self.get_result()


class BadRecordsPathSchemaOnRead(IValidator):
    def __init__(
        self,
        context,
        dataframe: DataFrame,
        exceptions_handler: Callable[[DataFrame], int],
        database: str,
        table: str,
        bad_records_path: str,
        spark: SparkSession,
    ) -> None:
        super().__init__(context, dataframe, exceptions_handler, database, table)
        self.path = bad_records_path
        self.spark = spark

    def validate(self):

        self.total_count = self.dataframe.count()
        self.dataframe.cache()
        self.valid_count = self.dataframe.distinct().count()
        self.exceptions_count = self.total_count - self.valid_count
        options = {INFER_SCHEMA: True, RECURSIVE_FILE_LOOKUP: True}
        try:
            self.context.log.info(
                f"{self.exceptions_count} schema on read exceptions found for dataset {self.table}"
            )
            if self.exceptions_count > 0:
                self.context.log.info(
                    f"Try loading {self.exceptions_count} exceptions for dataset {self.table} from {self.path}"
                )
                exceptions = (
                    self.spark.read.format(Format.JSON.name.lower())
                    .options(**options)
                    .load(self.path)
                    .withColumn(TIMESTAMP, fn.current_timestamp())
                    .withColumn(DATABASE, fn.lit(self.database))
                    .withColumn(TABLE, fn.lit(self.table))
                )
                self.exceptions_count = self.exceptions_handler(exceptions)
                self.context.log.info(
                    f"Deleting exceptions for dataset {self.table} from {self.path}"
                )
                self.context.fs.rm(self.path, True)

        except AnalysisException as e:
            if self.exceptions_count > 0:
                msg = f"There are {self.exceptions_count} exceptions but dataset for table {self.table} failed to load from path {self.table}"
                self.context.log.error(msg)
                raise Exception(msg) from e
            exceptions = None
            self.exceptions_count = 0

        return self.get_result()



import os
import shutil
from ._ifile_system import IFileSystem, FileFormat
from typing import Type, Union
import yaml
import json


class FileSystem(IFileSystem):
    def __init__(self, context: str, datalake_protocol: str = "file:") -> None:
        super().__init__(context, datalake_protocol)

    def rm(self, path: str, recurse=False) -> bool:
        """Removes a file or directory."""

        if os.path.exists(path):
            if os.path.isfile(path):
                os.remove(path)
                return True
            elif os.path.isdir(path):
                if recurse:
                    shutil.rmtree(path)
                    return True
                else:
                    os.rmdir(path)
                    return True
            else:
                return False
        else:
            return False

    def cp(self, from_path: str, to_path: str, recurse=False) -> bool:
        """Copies a file or directory, possibly across FileSystems."""

        if os.path.exists(from_path) and os.path.exists(to_path):
            if os.path.isfile(from_path):
                # TODO: shutil.copy need to add logic wherther destination is a folder or filename
                shutil.copyfile(from_path, to_path)
                return True
            elif os.path.isdir(from_path):
                if recurse:
                    shutil.copytree(from_path, to_path)
                    return True
                else:
                    # TODO: handle if it's directory that must have recurse true
                    False
            else:
                return False
        else:
            return False

    def mv(self, from_path: str, to_path: str, recurse=False) -> bool:
        """Moves a file or directory, possibly across FileSystems."""
        if not os.path.exists(to_path):
            self.mkdirs(to_path)

        if os.path.exists(from_path) and os.path.exists(to_path):
            shutil.move(from_path, to_path)
            return True
        else:
            return False

    def ls(self, path: str) -> list:
        """Copies a file or directory, possibly across FileSystems."""
        return os.listdir(path)

    def put(self, file: str, contents: str, overwrite=False) -> bool:
        """Writes the given String out to a file, encoded in UTF-8."""
        if overwrite:
            mode = "w"
        else:
            mode = "a"
        with open(file, mode) as f:
            f.write(contents)

    def head(self, file: str, maxBytes: int = 65536) -> str:
        """Returns up to the first 'maxBytes' bytes of the given file as a String encoded in UTF-8"""
        with open(file, "rb") as fp:
            try:
                data = fp.read(maxBytes)
            except EOFError:
                pass

        return data.decode("utf-8")

    def mkdirs(self, path: str) -> bool:
        """Creates the given directory if it does not exist, also creating any necessary parent directories"""
        if os.path.exists(path):
            return False
        else:
            os.makedirs(path)
            return True

    def read_file(self, path: str, file_format: FileFormat) -> str:

        with open(path, "r") as f:
            if file_format == FileFormat.JSON:
                data = json.loads(f.read())
            elif file_format == FileFormat.YAML:
                data = f.read()
                data = yaml.safe_load(data)
            elif file_format == FileFormat.TEXT:
                data = f.read()
            else:
                raise Exception(
                    f"File format not supported {file_format} when reading file {path}"
                )

        return data

    def append_file(self, path: str, data: Union[str, dict], file_format: FileFormat):

        if isinstance(data, dict) and file_format == FileFormat.TEXT:
            raise TypeError()
        if isinstance(data, str) and file_format in [
            FileFormat.JSON,
            FileFormat.YAML,
            FileFormat.JSONL,
        ]:
            raise TypeError()
        if not isinstance(data, (str, dict)):
            raise TypeError()

        with open(path, "a") as f:
            if file_format == FileFormat.JSON:
                data_formatted = json.dumps(data, indent=4, default=str)
                f.write(data_formatted)
            elif file_format == FileFormat.YAML:
                data_formatted = yaml.safe_dump(data, indent=4)
                f.write(data_formatted)
            elif file_format == FileFormat.TEXT:
                f.write(data)
            else:
                raise Exception(
                    f"File format not supported {file_format} when appending file {path}"
                )

    def write_file(self, path: str, data: Union[str, dict], file_format: FileFormat):

        if isinstance(data, dict) and file_format == FileFormat.TEXT:
            raise TypeError()
        if isinstance(data, str) and file_format in [
            FileFormat.JSON,
            FileFormat.YAML,
            FileFormat.JSONL,
        ]:
            raise TypeError()
        if not isinstance(data, (str, dict)):
            raise TypeError()

        with open(path, "w") as f:
            if file_format == FileFormat.JSON:
                data_formatted = json.dumps(data, indent=4, default=str)
                f.write(data_formatted)
            elif file_format == FileFormat.YAML:
                data_formatted = yaml.safe_dump(data, indent=4)
                f.write(data_formatted)
            elif file_format == FileFormat.TEXT:
                f.write(data)
            else:
                raise Exception(
                    f"File format not supported {file_format} when writing file {path}"
                )

    def exists(self, path: str) -> bool:
        return os.path.exists(path)



from ._imetadata_repo import IMetadataRepo
from uuid import UUID
from ..dataset import Dataset
from datetime import datetime
from ..file_system import FileFormat
from collections import ChainMap
import json


class MetadataFile(IMetadataRepo):

    _DEFAULT_DATASET_FILENAME = "dataset"
    _DEFAULT_INDEX_FILENAME = "index"
    _DEFAULT_EXT = FileFormat.JSON
    _DEFUALT_ROOT = "./config/runs"

    _ROOT = "metadata_root"
    _DATASET = "metadata_dataset"
    _INDEX = "metadata_index"

    def __init__(self, context, config: dict) -> None:
        super().__init__(context, config)
        _config = config["metadata_file"]
        self.root = _config.get(self._ROOT, self._DEFUALT_ROOT)
        self.dataset_filename = _config.get(
            self._DATASET,
            f"{self._DEFAULT_DATASET_FILENAME}.{self._DEFAULT_EXT.name.lower()}",
        )
        self.index_filename = _config.get(
            self._INDEX,
            f"{self._DEFAULT_INDEX_FILENAME}.{self._DEFAULT_EXT.name.lower()}",
        )

    def _merge_existing(
        self, metadata: dict, context_id: UUID, dir_path: str, file_path: str
    ):
        """If a file doesn't exist write a new data file. If a file does exist the read the contents
        merge with new data and write the file. Because at the moment it's a file per dataflow
        we just re-write the file since it won't be that big.

        """
        str_context_id = str(context_id)
        if not self.context.fs.exists(dir_path):
            self.context.fs.mkdirs(dir_path)

        if not self.context.fs.exists(file_path):
            metadata = {str_context_id: metadata}

        else:
            existing_metadata: dict = self.context.fs.read_file(
                file_path, FileFormat.JSON
            )

            metadata = ChainMap(existing_metadata[str_context_id], metadata)
            metadata = {str_context_id: dict(metadata)}

        return metadata

    def save(self, dataset: Dataset):

        dir_path = f"{self.root}/{dataset.context_id}"
        dataset_file_path = f"{dir_path}/{self.dataset_filename}"
        index_file_path = f"{dir_path}/{self.index_filename}"

        if not self.context.fs.exists(dir_path):
            self.context.fs.mkdirs(dir_path)

        metadata_dataset: dict = self._get_dataset(dataset)
        metadata_index: dict = self._get_index(dataset)

        metadata_dataset = self._merge_existing(
            metadata_dataset, dataset.context_id, dir_path, dataset_file_path
        )
        self.context.log.debug(json.dumps(metadata_dataset, indent=4, default=str))
        self.context.fs.write_file(dataset_file_path, metadata_dataset, FileFormat.JSON)

        metadata_index = self._merge_existing(
            metadata_index, dataset.context_id, dir_path, index_file_path
        )
        self.context.log.debug(json.dumps(metadata_index, indent=4, default=str))
        self.context.fs.write_file(index_file_path, metadata_index, FileFormat.JSON)



from enum import Enum


from ._spark_file_schema_repo import SparkFileSchemaRepo
from ._deltalake_sql_file import DeltalakeSchemaFile
from ._ischema_repo import ISchemaRepo
import logging


class SchemaRepoType(Enum):
    SPARK_SCHEMA_FILE = 1
    DELTALAKE_SQL_FILE = 2


class _SchemaRepoFactory:
    def __init__(self) -> None:
        self._logger = logging.getLogger(__name__)
        self._schema_repo = {}

    def register_schema_repo_type(
        self, sr_type: SchemaRepoType, schema_repo_type: type
    ):
        self._logger.debug(f"Register file system type {schema_repo_type} as {type}")
        self._schema_repo[sr_type] = schema_repo_type

    def _get_sr_type(self, name: str):
        name = name.strip().upper()
        try:
            if SchemaRepoType[name] in SchemaRepoType:
                return SchemaRepoType[name]
        except:
            return None

    def get_schema_repo_type(self, context, config: dict) -> ISchemaRepo:

        schema_repo_store: str = next(iter(config))
        sr_type: SchemaRepoType = self._get_sr_type(schema_repo_store)

        context.log.info(f"Setting up schema repo on {schema_repo_store} ")

        context.log.debug(f"Setting SchemaRepoType using type {sr_type}")
        schema_repo: ISchemaRepo = self._schema_repo.get(sr_type)

        if not schema_repo:
            self._logger.error(
                f"SchemaRepoType {sr_type.name} not registered in the schema_repo factory"
            )
            raise ValueError(sr_type)

        return schema_repo(context, config)


factory = _SchemaRepoFactory()
factory.register_schema_repo_type(SchemaRepoType.SPARK_SCHEMA_FILE, SparkFileSchemaRepo)
factory.register_schema_repo_type(
    SchemaRepoType.DELTALAKE_SQL_FILE, DeltalakeSchemaFile
)



import json
from ._ischema_repo import ISchemaRepo
from pyspark.sql.types import StructType
from ..file_system import FileFormat, IFileSystem, file_system_factory, FileSystemType


class DeltalakeSchemaFile(ISchemaRepo):

    _SCHEMA_ROOT = "./config/schema/deltalake"
    _EXT = "sql"

    def __init__(self, context, config: dict) -> None:
        super().__init__(context, config)
        self.root_path = config["deltalake_sql_file"].get("deltalake_schema_root")

    def _mkpath(self, database_name: str, table_name: str):
        """Function that builds the schema path"""
        if not self.root_path:
            return f"{self._SCHEMA_ROOT}/{database_name}/{table_name}.{self._EXT}"
        else:
            return f"{self.root_path}/{database_name}/{table_name}.{self._EXT}"

    def save_schema(self, schema: StructType, database_name: str, table_name: str):
        """Serialise delta table to a create table sql file."""
        raise NotImplementedError

    def load_schema(self, database_name: str, table_name: str):
        """Loads a spark from a yaml file and deserialises to a spark schema."""

        path = self._mkpath(database_name, table_name)

        # this was in thought that schema's could be maintain and loaded off DBFS for databricks
        # it actually works much better using the local repo files
        # maybe considered for a future use case - we need more configuration to set the schema store
        # type, since it's explicitly hand in hand with databricks spark env.
        # fs: IFileSystem = self.context.fs

        # file system is working fine for databricks and vanilla spark deployments.
        fs: IFileSystem = file_system_factory.get_file_system_type(
            self.context, FileSystemType.FILE
        )

        self.context.log.info(
            f"Loading schema for dataset {database_name}.{table_name} from {path} using {type(fs)}"
        )

        schema = fs.read_file(path, FileFormat.TEXT)
        if not schema:
            msg = f"Failed to load schema for dataset {database_name}.{table_name} from {path}"
            raise Exception(msg)

        msg = json.dumps(schema)
        self.context.log.debug(msg)

        return schema



from ._imetadata_repo import IMetadataRepo
from uuid import UUID
from ..dataset import Dataset


class MetadataDeltalake(IMetadataRepo):

    _DEFAULT_DATASET_TABLE = "dataset"
    _DEFAULT_INDEX_TABLE = "index"
    _DEFUALT_DATABSE = "yetl"

    _DATABASE = "metadata_database"
    _DATASET = "metadata_table"
    _INDEX = "metadata_index"

    def __init__(self, context, config: dict) -> None:

        super().__init__(context, config)
        _config = config["metadata_deltalake"]

        self.database = _config.get(self._ROOT, self._DEFUALT_DATABSE)
        self.dataset_table = _config.get(self._DATASET, self._DEFAULT_DATASET_TABLE)
        self.index_table = _config.get(self._INDEX, self._DEFAULT_INDEX_TABLE)

    def save(self, dataset: Dataset):
        return super().save(dataset)



from enum import Enum
from typing import Type
from ._dataset import Dataset
from ._save import Save, DefaultSave, AppendSave
from ._reader import Reader
from ._writer import Writer
from ._stream_reader import StreamReader
from ._stream_writer import StreamWriter
import logging


class IOType(Enum):
    READ = 1
    WRITE = 2
    READSTREAM = 3
    WRITESTREAM = 4


class _DatasetFactory:
    def __init__(self) -> None:
        self._logger = logging.getLogger(__name__)
        self._dataset = {}

    def register_dataset_type(self, io_type: IOType, dataset_type: type):
        self._logger.debug(f"Register dataset type {dataset_type} as {type}")
        self._dataset[io_type] = dataset_type

    def _get_io_type(self, name: str):
        try:
            if IOType[name.upper()] in IOType:
                return IOType[name.upper()]
        except:
            return None

    def get_dataset_type(
        self,
        context,
        database: str,
        table: str,
        dataset_config: dict,
        save_type: Type[Save] = DefaultSave,
    ) -> Dataset:

        type: IOType = next(
            iter(
                [
                    self._get_io_type(k)
                    for k in dataset_config.keys()
                    if self._get_io_type(k)
                ]
            )
        )

        self._logger.info(f"Get {type.name} from factory dataset")
        dataset_class = self._dataset.get(type)

        # TODO: fix this - need to register save types and lose the if
        if type == IOType.WRITE:

            class InjectDestination(dataset_class, save_type):
                pass

            dataset_class = InjectDestination

        if not dataset_class:
            self._logger.error(
                f"IOType {type.name} not registered in the dataset factory dataset"
            )
            raise ValueError(type)

        return dataset_class(
            context, database, table, dataset_config, type.name.lower()
        )


factory = _DatasetFactory()
factory.register_dataset_type(IOType.READ, Reader)
factory.register_dataset_type(IOType.WRITE, Writer)
factory.register_dataset_type(IOType.READSTREAM, StreamReader)
factory.register_dataset_type(IOType.WRITESTREAM, StreamWriter)



from ._dbfs_file_system import DbfsFileSystem
from ._file_system import FileSystem
from ._ifile_system import IFileSystem, FileFormat
from ._factory import factory as file_system_factory, FileSystemType

__all__ = [
    "DbfsFileSystem",
    "FileSystem",
    "IFileSystem",
    "file_system_factory",
    "FileFormat",
    "FileSystemType",
]



from enum import Enum
from abc import ABC
from pyspark.sql import DataFrame


class SaveMode(Enum):
    """https://spark.apache.org/docs/latest/sql-data-sources-load-save-functions.html#save-modes"""

    DEFAULT = "default"
    ERROR_IF_EXISTS = "errorifexists"
    APPEND = "append"
    OVERWRITE = "overwrite"
    IGNORE = "ignore"
    MERGE = "merge"
    OVERWRITE_SCHEMA = "overwriteSchema"


class Save(ABC):
    def __init__(self) -> None:
        pass

    def write(self):
        pass


class DefaultSave(Save):
    def write(self):
        self.context.log.info(
            f"Writer saving using the DefaultSave which is the configured save = {self.mode}."
        )
        (
            self.dataframe.write.format(self.format)
            .options(**self.options)
            .mode(self.mode)
            .partitionBy(*self.partitions)
            .save(self.path)
        )


class ErrorIfExistsSave(Save):
    def write(self):
        self.context.log.info(
            f"Writer saving using the {self.__class__.__name__} which is an injected save."
        )
        (
            self.dataframe.write.format(self.format)
            .options(**self.options)
            .mode(SaveMode.ERROR_IF_EXISTS.value)
            .partitionBy(*self.partitions)
            .save(self.path)
        )


class AppendSave(Save):
    def write(self):
        self.context.log.info(
            f"Writer saving using the {self.__class__.__name__} which is an injected save."
        )
        (
            self.dataframe.write.format(self.format)
            .options(**self.options)
            .mode(SaveMode.APPEND.value)
            .partitionBy(*self.partitions)
            .save(self.path)
        )


class OverwriteSchemaSave(Save):
    def write(self):
        self.context.log.info(
            f"Writer saving using the {self.__class__.__name__} which is an injected save."
        )
        options = self.options
        options[SaveMode.OVERWRITE_SCHEMA.value] = True
        (
            self.dataframe.write.format(self.format)
            .options(**options)
            .mode(SaveMode.OVERWRITE.value)
            .partitionBy(*self.partitions)
            .save(self.path)
        )


class OverwriteSave(Save):
    def write(self):
        self.context.log.info(
            f"Writer saving using the {self.__class__.__name__} which is an injected save."
        )
        (
            self.dataframe.write.format(self.format)
            .options(**self.options)
            .mode(SaveMode.OVERWRITE.value)
            .partitionBy(*self.partitions)
            .save(self.path)
        )


class IgnoreSave(Save):
    def write(self):
        self.context.log.info(
            f"Writer saving using the {self.__class__.__name__} which is an injected save."
        )
        (
            self.dataframe.write.format(self.format)
            .options(**self.options)
            .mode(SaveMode.IGNORE.value)
            .partitionBy(*self.partitions)
            .save(self.path)
        )


class MergeSave(Save):
    def write(self):
        self.context.log.info(
            f"Writer saving using the {self.__class__.__name__} which is an injected save."
        )
        raise NotImplementedError()



import re
from datetime import datetime
from tracemalloc import start
import regex
from ._constants import NAME, REPLACE, ARGS


class TimeslicePosition:
    def __init__(
        self, start_pos: int, end_pos: int, length: int, format_code: str
    ) -> None:
        self.start_pos = start_pos
        self.end_pos = end_pos
        self.length = length
        self.format_code = format_code


def sql_partitioned_by(sql: str):

    white_space = re.compile(r"\s+")
    sql = re.sub(white_space, "", sql).lower()

    partitioned_by = r"partitionedby\((.*)\)"
    partitions = re.search(partitioned_by, sql)
    if partitions:
        partitions_clause = partitions.group()
        partitions_clause = partitions_clause.replace("(", "")
        partitions_clause = partitions_clause.replace(")", "")
        partitions_clause = partitions_clause.replace("partitionedby", "")
        partitions_lst = partitions_clause.split(",")
        return partitions_lst
    else:
        return None


def to_spark_format_code(py_format: str):
    """Convert python format codes to spark format codes

    https://spark.apache.org/docs/latest/sql-ref-datetime-pattern.html

    %d -> dd    - Day of the month as a zero-padded decimal number.
    %m -> MM    - Month as a zero-padded decimal number.
    %y -> yy    - Year without century as a zero-padded decimal number.
    %Y -> yyyy  - Year with century as a decimal number.
    %H -> HH    - Hour (24-hour clock) as a zero-padded decimal number.
    %M -> mm    - Minute as a zero-padded decimal number.
    %S -> s     - Second as a zero-padded decimal number.
    %f -> S     - Microsecond as a decimal number, zero-padded to 6 digits.
    %% -> %     - A literal '%' character.
    %j -> DD    - Day of the year as a zero-padded decimal number.

    NOT SUPPORTED - %U - Week number of the year (Sunday as the first day of the week) as a zero-padded decimal number. All days in a new year preceding the first Sunday are considered to be in week 0.
    NOT SUPPORTED - %W - Week number of the year (Monday as the first day of the week) as a zero-padded decimal number. All days in a new year preceding the first Monday are considered to be in week 0.

    NOT SUPPORTED - %c - Locale’s appropriate date and time representation.
    NOT SUPPORTED - %x - Locale’s appropriate date representation.
    NOT SUPPORTED - %X - Locale’s appropriate time representation.
    NOT SUPPORTED - %G - ISO 8601 year with century representing the year that contains the greater part of the ISO week (%V).
    NOT SUPPORTED - %u - ISO 8601 weekday as a decimal number where 1 is Monday.
    NOT SUPPORTED - %V - ISO 8601 week as a decimal number with Monday as the first day of the week. Week 01 is the week containing Jan 4.
    NOT SUPPORTED - %z - UTC offset in the form ±HHMM[SS[.ffffff]] (empty string if the object is naive).
    NOT SUPPORTED - %Z - Time zone name (empty string if the object is naive).
    NOT SUPPORTED - %I - Hour (12-hour clock) as a zero-padded decimal number.
    NOT SUPPORTED - %p - Locale’s equivalent of either AM or PM.
    NOT SUPPORTED - %b - Month as locale’s abbreviated name.
    NOT SUPPORTED - %B - Month as locale’s full name.
    NOT SUPPORTED - %a - Weekday as locale’s abbreviated name.
    NOT SUPPORTED - %A - Weekday as locale’s full name.
    NOT SUPPORTED - %w - Weekday as a decimal number, where 0 is Sunday and 6 is Saturday."""

    UNSUPPORTED_FORMAT_CODES = [
        "%U",
        "%W",
        "%c",
        "%x",
        "%X",
        "%G",
        "%u",
        "%V",
        "%z",
        "%Z",
        "%I",
        "%p",
        "%b",
        "%B",
        "%a",
        "%A",
        "%w",
    ]
    unsupported_codes = []
    for c in UNSUPPORTED_FORMAT_CODES:
        if c in py_format:
            unsupported_codes.append(c)

    if unsupported_codes:
        unsupported_codes = ",".join(unsupported_codes)
        raise Exception(
            f"The format contains the following unsupported format codes: {unsupported_codes}"
        )

    py_format = py_format.replace("%d", "dd")
    py_format = py_format.replace("%m", "MM")
    py_format = py_format.replace("%y", "yy")
    py_format = py_format.replace("%Y", "yyyy")
    py_format = py_format.replace("%H", "HH")
    py_format = py_format.replace("%M", "mm")
    py_format = py_format.replace("%S", "s")
    py_format = py_format.replace("%f", "S")
    py_format = py_format.replace("%", "")

    return py_format


def parse_functions(sentence: str):

    rgx_function = regex.compile(r"(?si)(?|{0}(.*?){1}|{1}(.*?){0})".format("{{", "}}"))
    functions = rgx_function.findall(sentence)
    funcs = []
    for f in functions:
        for fp in f:
            func: str = fp
            args = func[func.find("(") + 1 : func.find(")")]
            args = args.split(",")
            args = [a.strip() for a in args]
            function_name = func[: func.find("(")].strip()
            function = {NAME: function_name, REPLACE: func, ARGS: args}
            funcs.append(function)

    return funcs


def get_slice_position(path: str, dataset):

    parsed = parse_functions(path)
    start_pos = path.find("{")
    end_pos = -1
    timeslice_pos = None

    if start_pos > 0:
        dte_format_attribute = parsed[0]["args"][0]

        if hasattr(dataset, dte_format_attribute):
            format_str = getattr(dataset, dte_format_attribute)

            l = len(datetime.now().strftime(format_str))

            end_pos = start_pos + l
            format_str = to_spark_format_code(format_str)

            timeslice_pos = TimeslicePosition(start_pos + 3, end_pos + 3, l, format_str)

    return timeslice_pos



from ._source import Source


class StreamReader(Source):
    def __init__(
        self, context, database: str, table: str, dataset: dict, io_type: str
    ) -> None:
        super().__init__(context, database, table, dataset, io_type)



from ._dataset import Dataset
from ._source import Source
from ._destination import Destination
from ._reader import Reader
from ._writer import Writer
from ._stream_reader import StreamReader
from ._stream_writer import StreamWriter
from ._factory import factory as dataset_factory
from ._save import (
    AppendSave,
    IgnoreSave,
    OverwriteSave,
    OverwriteSchemaSave,
    ErrorIfExistsSave,
    MergeSave,
    DefaultSave,
    Save,
)

__all__ = [
    "Source",
    "Destination",
    "Dataset",
    "Reader",
    "Writer",
    "StreamReader",
    "StreamWriter",
    "dataset_factory",
    "AppendSave",
    "IgnoreSave",
    "OverwriteSave",
    "OverwriteSchemaSave",
    "ErrorIfExistsSave",
    "MergeSave",
    "DefaultSave",
    "Save",
]



from ._validation import PermissiveSchemaOnRead, BadRecordsPathSchemaOnRead
from pyspark.sql import DataFrame
import json
from ._dataset import Dataset
from ..parser._constants import (
    Format,
    MERGE_SCHEMA,
    APPEND,
    BAD_RECORDS_PATH,
    CORRUPT_RECORD,
)


class Source(Dataset):
    def __init__(
        self, context, database: str, table: str, dataset: dict, io_type: str
    ) -> None:
        super().__init__(context, database, table, dataset, io_type)
        self.dataframe: DataFrame = None
        self.auto_io: bool

    def validate(self):
        pass

    def read(self):
        pass

    def is_source(self):
        return True

    def is_destination(self):
        return False

    def save_metadata(self):
        self.context.metadata_repo.save(self)



from pyspark.sql import DataFrame, functions as fn
from ..dataset import Dataset, Source, Destination
from ._i_dataflow import IDataflow
from ..dataset import dataset_factory, Save, DefaultSave
from typing import Type
from enum import Enum


class DataFlowType(Enum):
    """The type of lead e.g. incremental, incremental CDC, full load etc
    FULL_LOAD - Selects all the data, has the option of supplying a slice * mask to the timeslice
    PARTITION_LOAD - Loads / reloads affected partitions
    MERGE_COMPARE - Merges based on a full comparison of change tracking fields
    MERGE_CDC - Merges based on CDC provided metadata
    NUKE - Drops all the tables and data before fully reloading everything
    """

    FULL_LOAD = (0,)
    PARTITION_LOAD = 1
    MERGE_COMPARE = (2,)
    MERGER_CDC = (3,)
    NUKE = 100


class Dataflow(IDataflow):
    def __init__(
        self,
        context,
        config: dict,
        dataflow_config: dict,
        save_type: Type[Save] = DefaultSave,
    ) -> None:

        super().__init__(context, config, dataflow_config, save_type)

        for database, v in dataflow_config.items():
            for table, v in v.items():
                v["datalake"] = self.datalake
                v["datalake_protocol"] = self.datalake_protocol
                v["spark_schema_repo"] = self._spark_schema_repo
                v["deltalake_schema_repo"] = self._deltalake_schema_repo
                v["context_id"] = self.context.context_id
                v["timeslice"] = self.context.timeslice
                md = dataset_factory.get_dataset_type(
                    self.context, database, table, v, save_type
                )
                self.log.debug(
                    f"Deserialized {database}.{table} configuration into {type(md)}"
                )
                self.append(md)

    def append(self, dataset: Dataset):

        if dataset.is_source():
            self.log.debug(
                f"Appending source {dataset.database_table} as {type(dataset)} to dataflow"
            )
            if dataset.auto_io:
                source: Source = dataset
                source.read()
            self.sources[dataset.database_table] = source

        elif dataset.is_destination():
            self.log.debug(
                f"Appending destination {dataset.database_table} as {type(dataset)} to dataflow"
            )
            self.destinations[dataset.database_table] = dataset

    def source_df(self, database_table: str):

        src: Source = self.sources[database_table]
        return src.dataframe

    def destination_df(self, database_table: str, dataframe: DataFrame):

        dst: Destination = self.destinations[database_table]
        dst.dataframe = dataframe
        if dst.auto_io:
            dst.write()



from enum import Enum


class FileFormat(Enum):
    TEXT = 1
    JSON = 2
    YAML = 3
    JSONL = 4


class IFileSystem:
    def __init__(self, context, datalake_protocol: str) -> None:
        self.context = context
        self.datalake_protocol = datalake_protocol

    def rm(self, path: str, recurse=False) -> bool:
        """Removes a file or directory."""
        pass

    def cp(self, from_path: str, to_path: str, recurse=False) -> bool:
        """Copies a file or directory, possibly across FileSystems."""
        pass

    def mv(self, from_path: str, to_path: str, recurse=False) -> bool:
        """Moves a file or directory, possibly across FileSystems."""
        pass

    def ls(self, path: str) -> list:
        """Copies a file or directory, possibly across FileSystems."""
        pass

    def put(self, file: str, contents: str, overwrite=False) -> bool:
        """Writes the given String out to a file, encoded in UTF-8."""
        pass

    def head(self, file: str, maxBytes: int = 65536) -> str:
        """Returns up to the first 'maxBytes' bytes of the given file as a String encoded in UTF-8"""
        pass

    def mkdirs(self, path: str) -> bool:
        """Creates the given directory if it does not exist, also creating any necessary parent directories"""
        pass

    def read_file(self, path: str, file_format: FileFormat) -> str:
        pass

    def append_file(self, path: str, data: dict, file_format: FileFormat):
        pass

    def write_file(self, path: str, data: dict, file_format: FileFormat):
        pass

    def exists(self, path: str) -> bool:
        pass



from ._dataset import Dataset
from pyspark.sql import DataFrame
from ._save import DefaultSave
from pyspark.sql import functions as fn
from ..parser._constants import *
import json


class Destination(Dataset):
    def __init__(
        self, context, database: str, table: str, dataset: dict, io_type: str
    ) -> None:
        super().__init__(context, database, table, dataset, io_type)

        self.auto_io: bool
        self.dataframe: DataFrame

    def write(self):

        self.context.log.debug(
            f"Reordering sys_columns to end for {self.database_table} from {self.path} {CONTEXT_ID}={str(self.context_id)}"
        )
        # remove a re-add the _context_id since there will be dupplicate columns
        # when dataframe is built from multiple sources.
        self.dataframe = self.dataframe.drop(CONTEXT_ID).withColumn(
            CONTEXT_ID, fn.lit(str(self.context_id))
        )

        # clean up the column ordering
        sys_columns = [c for c in self.dataframe.columns if c.startswith("_")]
        data_columns = [c for c in self.dataframe.columns if not c.startswith("_")]
        data_columns = data_columns + sys_columns
        self.dataframe = self.dataframe.select(*data_columns)

        # get the partitions values for efficient IO patterns
        self.partition_values = self._get_partitions_values()

        if self.partition_values:
            msg_partition_values = json.dumps(self.partition_values, indent=4)
            self.context.log.info(
                f"""IO operations for {self.database}.{self.table} will be paritioned by: \n{msg_partition_values}"""
            )

        super().write()
        self.save_metadata()

    def _get_partitions_values(self):

        partition_values = {}
        if self.partitions:
            partition_values_df = self.dataframe.select(*self.partitions).distinct()

            for p in self.partitions:
                group_by: list = list(self.partitions)
                group_by.remove(p)
                if group_by:
                    partition_values_df = partition_values_df.groupBy(*group_by).agg(
                        fn.collect_set(p).alias(p)
                    )
                else:
                    partition_values_df = partition_values_df.withColumn(
                        p, fn.collect_set(p)
                    )

            partition_values_df = partition_values_df.collect()
            partition_values = partition_values_df[0].asDict()

        return partition_values

    def is_source(self):
        return False

    def is_destination(self):
        return True

    def save_metadata(self):
        return super().save_metadata()



from enum import Enum

from ._dbfs_file_system import DbfsFileSystem
from ._file_system import FileSystem
from ._ifile_system import IFileSystem
from typing import Union
import logging


class FileSystemType(Enum):
    FILE = 1
    DBFS = 2


class _FileSystemFactory:
    def __init__(self) -> None:
        self._logger = logging.getLogger(__name__)
        self._file_system = {}

    def register_file_system_type(
        self, fs_type: FileSystemType, file_system_type: type
    ):
        self._logger.debug(f"Register file system type {file_system_type} as {type}")
        self._file_system[fs_type] = file_system_type

    def _get_fs_type(self, name: str):
        name = name.replace(":", "").strip().upper()
        try:
            if FileSystemType[name] in FileSystemType:
                return FileSystemType[name]
        except:
            return None

    def get_file_system_type(
        self, context, config: Union[dict, FileSystemType]
    ) -> IFileSystem:

        if isinstance(config, dict):

            # determine type from configuration

            datalake_protocol: str = config["datalake_protocol"]
            fs_type: FileSystemType = self._get_fs_type(datalake_protocol)
            if not fs_type:
                context.log.error(
                    f"FileSystemType {fs_type.name} not registered in the file_system factory"
                )
                raise ValueError(fs_type)

            context.log.info(f"Setting filestystem using protocol {datalake_protocol}")
            context.log.debug(f"Setting FileSystemType using type {fs_type}")
            file_system: IFileSystem = self._file_system.get(fs_type)
            return file_system(context, datalake_protocol)

        elif isinstance(config, FileSystemType):
            # return based on the type asked for.
            context.log.debug(f"Setting FileSystemType using type {config}")
            file_system: IFileSystem = self._file_system.get(config)
            return file_system(context)

        else:
            raise Exception(f"FileSystemType cannot be produced using {type(config)}")


factory = _FileSystemFactory()
factory.register_file_system_type(FileSystemType.FILE, FileSystem)
factory.register_file_system_type(FileSystemType.DBFS, DbfsFileSystem)



from enum import Enum

from ._metadata_deltalake import MetadataDeltalake
from ._metadata_file import MetadataFile
from ._imetadata_repo import IMetadataRepo
import logging


class MetadataRepoType(Enum):
    METADATA_FILE = 1
    METADATA_DATALAKE = 2


class _MetadataRepoFactory:
    def __init__(self) -> None:
        self._logger = logging.getLogger(__name__)
        self._metadata_repo = {}

    def register_metadata_repo_type(
        self, mr_type: MetadataRepoType, metadata_repo_type: type
    ):
        self._logger.debug(
            f"Register metadata repo type {metadata_repo_type} as {type}"
        )
        self._metadata_repo[mr_type] = metadata_repo_type

    def _get_mr_type(self, name: str):
        try:
            if MetadataRepoType[name.upper()] in MetadataRepoType:
                return MetadataRepoType[name.upper()]
        except:
            return None

    def get_metadata_repo_type(self, context, config: dict) -> IMetadataRepo:

        # default to file metadata repo
        _config = config.get("metadata_repo", "metadata_file")
        metadata_repo_store: str = next(iter(_config))
        mr_type: MetadataRepoType = self._get_mr_type(metadata_repo_store)

        context.log.info(f"Setting up metadata repo on {metadata_repo_store} ")

        context.log.debug(f"Setting MetadataRepoType using type {mr_type}")
        metadata_repo: IMetadataRepo = self._metadata_repo.get(mr_type)

        if not metadata_repo:
            self._logger.error(
                f"MetadataRepoType {mr_type.name} not registered in the metadata_repo factory"
            )
            raise ValueError(mr_type)

        return metadata_repo(context, _config)


factory = _MetadataRepoFactory()
factory.register_metadata_repo_type(MetadataRepoType.METADATA_FILE, MetadataFile)
factory.register_metadata_repo_type(
    MetadataRepoType.METADATA_DATALAKE, MetadataDeltalake
)



from doctest import FAIL_FAST
from ._source import Source
from pyspark.sql import functions as fn
from pyspark.sql.types import StructType
from ..parser._constants import *
from . import _builtin_functions as builtin_funcs
from ..schema_repo import ISchemaRepo, SchemaNotFound
from ._validation import PermissiveSchemaOnRead, BadRecordsPathSchemaOnRead
from pyspark.sql import DataFrame
import json
from .. import _delta_lake as dl
from ..parser import parser


class Reader(Source):
    def __init__(
        self, context, database: str, table: str, config: dict, io_type: str
    ) -> None:
        super().__init__(context, database, table, config, io_type)

        # gets the read, write, etc options based on the type
        io_properties = config.get(io_type)

        # get the table properties
        properties: dict = self._get_table_properties(config["table"])

        self._metadata_lineage_enabled = properties.get(
            YETL_TBLP_METADATA_LINEAGE_ENABLED, False
        )
        self._create_schema_if_not_exists = properties.get(
            YETL_TBLP_SCHEMA_CREATE_IF_NOT_EXISTS, False
        )
        self._metadata_timeslice_tabled = properties.get(
            YETL_TBLP_METADATA_TIMESLICE, True
        )

        self.has_exception_configured = EXCEPTIONS in config.keys()
        self.auto_io = io_properties.get(AUTO_IO, True)
        self.options: dict = io_properties.get(OPTIONS)
        self.infer_schema = self.options.get(INFER_SCHEMA, False)
        self.has_badrecordspath_configured = BAD_RECORDS_PATH in self.options.keys()
        self.mode = self.options.get(MODE, None)

        # try and load a schema if schema on read
        try:
            self.schema = self._get_schema(config["spark_schema_repo"])
            self.has_schema = True


        except SchemaNotFound as e:
            self.schema = None
            self.has_schema = False
            if not self.infer_schema and not self._create_schema_if_not_exists:
                self._creating_inferred_schema = False
                msg = f"schema not found for {self.database}.{self.table} as path {e}"
                raise SchemaNotFound
            else:
                msg = f"schema not found for {self.database}.{self.table} as path {e}"
                self.context.log.warning(msg)

        self._creating_inferred_schema = (
            self._create_schema_if_not_exists and not self.has_schema
        )

        if self.auto_io and self._creating_inferred_schema:
            self._on_schema_creation()

        if self.has_exception_configured:
            self._set_exceptions_attributes(config)

        if not self._creating_inferred_schema:
            self._validate_configuration()

        if self.has_badrecordspath_configured:
            self.options = self.configure_badrecords_path(
                self.options, self.datalake_protocol, self.datalake
            )

        # set the exceptions table settings if configured.

        self.database_table = f"{self.database}.{self.table}"
        # self._initial_load = super().initial_load

        if self.auto_io and self.has_exception_configured:
            self.context.log.info(
                f"auto_io = {self.auto_io} automatically creating or altering exception delta table {self.database}.{self.table} {CONTEXT_ID}={str(self.context_id)}"
            )
            self.create_or_alter_table()

    def _validate_configuration(self):

        if not self.has_badrecordspath_configured and not self.mode:
            self.context.log.warning(
                f"{BAD_RECORDS_PATH} and {MODE} option are not set, defaulting to {MODE}={FAIL_FAST} {CONTEXT_ID}={str(self.context_id)}"
            )
            self.options[MODE] = FAIL_FAST

        self.has_corrupt_column = self._is_corrupt_column_set(self.options, self.schema)

        if self.has_badrecordspath_configured and self.mode:
            msg = f"{BAD_RECORDS_PATH} and {MODE} option are both set, this is not supported. Configure either {BAD_RECORDS_PATH} or {MODE} {CONTEXT_ID}={str(self.context_id)}"
            self.context.log.error(msg)
            raise Exception(msg)

        if self.has_badrecordspath_configured and not self.has_schema:
            msg = f"No schema has been defined. {BAD_RECORDS_PATH} requires that a schema is defined. {CONTEXT_ID}={str(self.context_id)}"
            self.context.log.error(msg)
            raise Exception(msg)

        if self.has_badrecordspath_configured and not self.has_exception_configured:
            msg = f"{BAD_RECORDS_PATH} requires that {EXCEPTIONS} configuration is defined. {CONTEXT_ID}={str(self.context_id)}"
            self.context.log.error(msg)
            raise Exception(msg)

        if self.has_badrecordspath_configured and not self.has_corrupt_column:
            msg = f"{BAD_RECORDS_PATH} doesn't support the use of the _corrupt_record columnd. {CONTEXT_ID}={str(self.context_id)}"
            self.context.log.error(msg)
            raise Exception(msg)

        if self.mode and not self.has_schema:
            msg = f"No schema has been defined. {MODE}={self.mode} configuration requires that a schema is defined. {CONTEXT_ID}={str(self.context_id)}"
            self.context.log.error(msg)
            raise Exception(msg)

        if self.mode.lower()!=PERMISSIVE and self.has_exception_configured:
            msg = f"{MODE}={self.mode}, exceptions can only be handled on a mode={PERMISSIVE}, {EXCEPTIONS} configuration will be disabled. {CONTEXT_ID}={str(self.context_id)}"
            self.context.log.warning(msg)
            self.has_exception_configured = False

        if self.mode.lower()==PERMISSIVE and self.has_exception_configured and not self.has_corrupt_column:
            msg = f"If expceptions are configured for {MODE}={self.mode} then _corrupt_record columns must be supplied in the schema. {CONTEXT_ID}={str(self.context_id)}"
            self.context.log.error(msg)
            raise Exception(msg)

    def _on_schema_creation(self):

        self.initial_load = True
        self.has_corrupt_column = False

        if self.has_exception_configured:
            msg = f"Creating inferred schema for {self.database}.{self.table} {EXCEPTIONS} configuration will be ignored"
            self.context.log.warning(msg)
            self.has_exception_configured = False

        if self.has_badrecordspath_configured:
            msg = f"Creating inferred schema for {self.database}.{self.table} {BAD_RECORDS_PATH} configuration will be ingnored"
            self.context.log.warning(msg)
            self.has_badrecordspath_configured = False

        if self.mode:
            msg = f"Creating inferred schema for {self.database}.{self.table} {self.mode} configuration will be ignored"
            self.context.log.warning(msg)
            del self.options[MODE]

        if not self.infer_schema:
            msg = f"Creating inferred schema for {self.database}.{self.table} defaulting {INFER_SCHEMA} to True"
            self.context.log.warning(msg)
            self.options[INFER_SCHEMA] = True

    def _get_table_properties(self, table_config: dict):
        return table_config.get(PROPERTIES, {})

    def create_or_alter_table(self):

        dl.create_database(self.context, self.exceptions_database)
        table_exists = dl.table_exists(
            self.context, self.exceptions_database, self.exceptions_table
        )
        if table_exists:
            self.context.log.info(
                f"Exception table already exists {self.exceptions_database}.{self.exceptions_table} at {self.exceptions_path} {CONTEXT_ID}={str(self.context_id)}"
            )
            self.initial_load = False
        else:
            dl.create_table(
                self.context,
                self.exceptions_database,
                self.exceptions_table,
                self.exceptions_path,
            )
            self.initial_load = True

    @property
    def initial_load(self):

        return self._initial_load

    @initial_load.setter
    def initial_load(self, value: bool):

        self._initial_load = value

    def _get_schema(self, config: dict):

        self.schema_repo: ISchemaRepo = (
            self.context.schema_repo_factory.get_schema_repo_type(self.context, config)
        )
        schema = self.schema_repo.load_schema(self.database, self.table)
        return schema

    def _save_schema(self, schema: StructType):
        schema = self.schema_repo.save_schema(schema, self.database, self.table)

    def configure_badrecords_path(
        self, options: dict, datalake_protocol: str, datalake: str
    ):

        bad_records_path = options[BAD_RECORDS_PATH]
        bad_records_path = builtin_funcs.execute_replacements(bad_records_path, self)
        options[BAD_RECORDS_PATH] = f"{datalake_protocol}{datalake}/{bad_records_path}"

        return options

    def _set_exceptions_attributes(self, dataset: dict):

        exceptions = dataset[EXCEPTIONS]
        self.exceptions_table = exceptions.get(TABLE)
        self.exceptions_database = exceptions.get(DATABASE)
        self.exceptions_database_table = (
            f"{self.exceptions_database}.{self.exceptions_table}"
        )
        exceptions_path = exceptions.get(PATH)
        self.exceptions_path = f"{self.datalake_protocol}{self.datalake}/{exceptions_path}/{self.exceptions_database}/{self.exceptions_table}"

    def _is_corrupt_column_set(self, options: dict, schema: StructType):

        if schema:
            has_corrupt_column = (
                options.get(MODE, "").lower() == PERMISSIVE
                and CORRUPT_RECORD in schema.fieldNames()
            )

            if (
                options.get(MODE, "").lower() == PERMISSIVE
                and not CORRUPT_RECORD in schema.fieldNames()
            ):
                self.context.log.warning(
                    f"mode={PERMISSIVE} and corrupt record is not set in the schema, schema on read corrupt records will be silently dropped."
                )
        else:
            has_corrupt_column = False

        return has_corrupt_column

    def _get_validation_exceptions_handler(self):
        def handle_exceptions(exceptions: DataFrame):
            exceptions_count = 0
            if self.has_exception_configured:
                exceptions_count = exceptions.count()
                if exceptions and exceptions_count > 0:
                    options = {MERGE_SCHEMA: True}
                    self.context.log.warning(
                        f"Writing {exceptions_count} exception(s) from {self.database_table} to {self.exceptions_database_table} delta table {CONTEXT_ID}={str(self.context_id)}"
                    )
                    exceptions.write.format(Format.DELTA.value).options(**options).mode(
                        APPEND
                    ).save(self.exceptions_path)
            return exceptions_count

        return handle_exceptions

    def validate(self):

        validation_handler = self._get_validation_exceptions_handler()
        validator = None

        if self.has_badrecordspath_configured:

            bad_records_path = self.options[BAD_RECORDS_PATH]
            self.context.log.info(
                f"Validating dataframe read using badRecordsPath at {bad_records_path} {CONTEXT_ID}={str(self.context_id)}"
            )
            validator = BadRecordsPathSchemaOnRead(
                self.context,
                self.dataframe,
                validation_handler,
                self.database,
                self.table,
                bad_records_path,
                self.context.spark,
            )

        if self.has_corrupt_column:
            self.context.log.info(
                f"Validating dataframe read using PERMISSIVE corrupt column at {CORRUPT_RECORD} {CONTEXT_ID}={str(self.context_id)}"
            )
            validator = PermissiveSchemaOnRead(
                self.context,
                self.dataframe,
                validation_handler,
                self.database,
                self.table,
            )

        if validator:
            validation = validator.validate()
            self.dataframe = validator.dataframe
            self.context.log.info(json.dumps(validation, indent=4, default=str))

    def read(self):
        self.context.log.info(
            f"Reading data for {self.database_table} from {self.path} with options {self.options} {CONTEXT_ID}={str(self.context_id)}"
        )

        self.context.log.debug(json.dumps(self.options, indent=4, default=str))

        input_file_name = f"_path_{self.database}_{self.table}"

        df = self.context.spark.read.format(self.format)

        if self.has_schema:
            df = df.schema(self.schema)

        df: DataFrame = (
            df.options(**self.options)
            .load(self.path)
            .withColumn(CONTEXT_ID, fn.lit(str(self.context_id)))
        )

        if self._metadata_timeslice_tabled:
            df: DataFrame = (
                df
                # TODO .withColumn(input_file_name, fn.input_file_name())
                .withColumn(TIMESLICE, fn.input_file_name())
                .withColumn(
                    TIMESLICE,
                    fn.substring(
                        TIMESLICE,
                        self._timeslice_position.start_pos,
                        self._timeslice_position.length,
                    ),
                )
                .withColumn(
                    TIMESLICE,
                    fn.to_timestamp(TIMESLICE, self._timeslice_position.format_code),
                )
            )

        # if there isn't a schema and it's configured to create one the save it to repo.
        if self._creating_inferred_schema:
            self.context.log.info(f"Saving inferred schema for {self.database}.{self.table} into schema repository. {CONTEXT_ID}={str(self.context_id)}")
            self.schema = df.schema
            self.schema_repo.save_schema(self.schema, self.database, self.table)

        self.context.log.debug(
            f"Reordering sys_columns to end for {self.database_table} from {self.path}. {CONTEXT_ID}={str(self.context_id)}"
        )
        self.dataframe = df
        self.validation_result = self.validate()
        self.save_metadata()
        return self.dataframe

    def save_metadata(self):
        super().save_metadata()



import logging
import logging.config
import yaml
import os

from yaml import YAMLError


"""Returns as of a named logger based on a yaml logging configuration file

The configuration file is called logging.yaml. It's a directory
location is taken from an envrionment variables called DATAPREPHOME.
If this does not exist then it is defaulted to ./config/
"""

log_config_file = os.getenv("YETL_HOME", "./config/")
log_config_file = f"{log_config_file}logging.yaml"
log_config_file = os.path.abspath(log_config_file)

# check that it exists
if not os.path.exists(log_config_file):
    msg = f"Config logging file path does not exist {log_config_file}"
    raise Exception(msg)

# load the logging configuration into the logger
with open(log_config_file, "r") as f:

    try:
        config = yaml.safe_load(f.read())
        logging.config.dictConfig(config)

        # if it errors because of invalid yaml format then
        # provide details so the users can easily find and correct
        # if it's a different exception just let it raise
    except YAMLError as e:

        location = ""
        if hasattr(e, "problem_mark"):
            mark = e.problem_mark
            location = f"Error position ({mark.line}, {mark.column})"

        if hasattr(e, "problem"):
            problem = f"{e.problem}."

        raise Exception(
            f"Invalid yaml format in {log_config_file}. {problem} {location}"
        )



from pyspark.sql import SparkSession
from ._ifile_system import IFileSystem, FileFormat
from ._ifile_system import IFileSystem, FileFormat
from typing import Union
import yaml
import json


class DbfsFileSystem(IFileSystem):
    def __init__(self, context: str, datalake_protocol: str = "dbfs:") -> None:
        super().__init__(context, datalake_protocol)
        self._fs = self._get_dbutils(context.spark).fs

    def _get_dbutils(self, spark: SparkSession):
        from pyspark.dbutils import DBUtils

        return DBUtils(spark)

    def rm(self, path: str, recurse=False) -> bool:
        """Removes a file or directory."""
        return self._fs.rm(path, recurse)

    def cp(self, from_path: str, to_path: str, recurse=False) -> bool:
        """Copies a file or directory, possibly across FileSystems."""
        return self._fs.cp(from_path, to_path, recurse)

    def mv(self, from_path: str, to_path: str, recurse=False) -> bool:
        """Moves a file or directory, possibly across FileSystems."""
        return self._fs.mv(from_path, to_path, recurse)

    def ls(self, path: str) -> list:
        """Copies a file or directory, possibly across FileSystems."""
        return self._fs.ls(path)

    def put(self, file: str, contents: str, overwrite=False) -> bool:
        """Writes the given String out to a file, encoded in UTF-8."""
        return self._fs.put(file, contents, overwrite)

    def head(self, file: str, maxBytes: int = 65536) -> str:
        """Returns up to the first 'maxBytes' bytes of the given file as a String encoded in UTF-8"""
        return self._fs.head(file, maxBytes)

    def mkdirs(self, path: str) -> bool:
        """Creates the given directory if it does not exist, also creating any necessary parent directories"""
        return self._fs.mkdirs(path)

    def read_file(self, path: str, file_format: FileFormat) -> str:

        if path[0] != ".":
            path = f"/dbfs{path}"
        with open(path, "r") as f:
            if file_format == FileFormat.JSON:
                data = json.loads(f.read())
            elif file_format == FileFormat.YAML:
                data = f.read()
                data = yaml.safe_load(data)
            elif file_format == FileFormat.TEXT:
                data = f.read()
            else:
                raise Exception(
                    f"File format not supported {file_format} when reading file {path}"
                )

        return data

    def append_file(self, path: str, data: Union[str, dict], file_format: FileFormat):

        if path[0] != ".":
            path = f"/dbfs{path}"
        if isinstance(data, dict) and file_format == FileFormat.TEXT:
            raise TypeError()
        if isinstance(data, str) and file_format in [
            FileFormat.JSON,
            FileFormat.YAML,
            FileFormat.JSONL,
        ]:
            raise TypeError()
        if not isinstance(data, (str, dict)):
            raise TypeError()

        with open(path, "a") as f:
            if file_format == FileFormat.JSON:
                data_formatted = json.dumps(data, indent=4, default=str)
                f.write(data_formatted)
            elif file_format == FileFormat.YAML:
                data_formatted = yaml.safe_dump(data, indent=4)
                f.write(data_formatted)
            elif file_format == FileFormat.TEXT:
                f.write(data)
            else:
                raise Exception(
                    f"File format not supported {file_format} when appending file {path}"
                )

    def write_file(self, path: str, data: Union[str, dict], file_format: FileFormat):

        if path[0] != ".":
            path = f"/dbfs{path}"
        if isinstance(data, dict) and file_format == FileFormat.TEXT:
            raise TypeError()
        if isinstance(data, str) and file_format in [
            FileFormat.JSON,
            FileFormat.YAML,
            FileFormat.JSONL,
        ]:
            raise TypeError()
        if not isinstance(data, (str, dict)):
            raise TypeError()

        with open(path, "w") as f:
            if file_format == FileFormat.JSON:
                data_formatted = json.dumps(data, indent=4, default=str)
                f.write(data_formatted)
            elif file_format == FileFormat.YAML:
                data_formatted = yaml.safe_dump(data, indent=4)
                f.write(data_formatted)
            elif file_format == FileFormat.TEXT:
                f.write(data)
            else:
                raise Exception(
                    f"File format not supported {file_format} when writing file {path}"
                )



import pathlib
from setuptools import setup, find_packages


# The directory containing this file
HERE = pathlib.Path(__file__).parent

# The text of the README file
README = (HERE / "README.md").read_text()

# This call to setup() does all the work
setup(
    name="yetl-framework",
    version="0.0.13",
    description="yet (another spark) etl framework",
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://www.yetl.io/",
    project_urls={
        "GitHub": "https://github.com/sibytes/yetl",
        "Documentation": "https://www.yetl.io/",
    },
    license="MIT",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
    ],
    packages=find_packages(),
    install_requires=["pyyaml", "pyspark", "regex", "delta-spark"],
    zip_safe=False,
)



from ._metadata_file import MetadataFile
from ._metadata_deltalake import MetadataDeltalake
from ._imetadata_repo import IMetadataRepo
from ._factory import factory as metadata_repo_factory

__all__ = [
    "IMetadataRepo",
    "MetadataFile",
    "MetadataDeltalake",
    "metadata_repo_factory",
]



from .dataflow import Dataflow
from . import _config_provider as cp
from pyspark.sql import SparkSession
from datetime import datetime
from .file_system import file_system_factory, IFileSystem
import logging
import uuid
from .schema_repo import schema_repo_factory
import json
from ._timeslice import Timeslice, TimesliceUtcNow
from .dataset import Save, DefaultSave
from typing import Type
from delta import configure_spark_with_delta_pip
from .metadata_repo import metadata_repo_factory, IMetadataRepo


class Context:
    def __init__(
        self,
        app_name: str,
        log_level: str,
        name: str,
        spark: SparkSession = None,
        timeslice: datetime = None,
        save_type: Type[Save] = DefaultSave,
    ) -> None:
        self.context_id = uuid.uuid4()
        self.name = name
        self.app_name = app_name

        if not app_name:
            self.app_name = self.name

        self.timeslice: Timeslice = timeslice
        if not self.timeslice:
            self.timeslice = TimesliceUtcNow()

        self.log = logging.getLogger(self.app_name)
        self.log_level = log_level
        self.spark = spark

        # load application configuration
        config: dict = cp.load_config(self.app_name)

        if not spark:
            self.spark = self._get_spark_context(app_name, config)

        # set up the spark logger, the application has a python logger built in
        # but we also make the spark logger available should it be needed
        # because the spark logger is extremely verbose it's useful to the able
        # to set the level and use python native logging.
        self.log.info(f"Setting application context spark logger at level {log_level}")
        self.spark_logger = self._get_spark_logger(
            self.spark, self.app_name, self.log_level
        )

        # abstraction of the filesystem for driver file commands e.g. rm, ls, mv, cp
        self.fs: IFileSystem = file_system_factory.get_file_system_type(
            self, config=config
        )

        # abstraction of the metadata repo for saving yetl dataflow lineage.
        self.metadata_repo: IMetadataRepo = (
            metadata_repo_factory.get_metadata_repo_type(self, config=config)
        )

        # abstraction of the schema repo
        self.schema_repo_factory = schema_repo_factory

        # Load and deserialise the spark dataflow configuration in to metaclasses (see dataset module)
        # The configuration file is loaded using the app name. This keeps intuitive tight
        # naming convention between datadlows and the config files that store them
        self.log.info(f"Setting application context dataflow {self.name}")
        self.dataflow = self._get_deltalake_flow(
            self.app_name, self.name, config, save_type
        )

        self.log.info(f"Checking spark and databricks versions")
        self.spark_version, self.databricks_version = self._get_spark_version(
            self.spark
        )

        if self.databricks_version:
            self.is_databricks = True
            self.log.info(
                f"Databricks Runtime version detected as : {self.databricks_version}"
            )
        else:
            self.is_databricks = False
            self.log.info(f"Databricks Runtime version not detected.")

        self.log.info(f"Spark version detected as : {self.spark_version}")

    def _get_spark_version(self, spark: SparkSession):

        version: str = spark.sql("select version() as version").collect()[0]["version"]

        try:
            databricks_version: dict = (
                spark.sql("select current_version() as version")
                .collect()[0]["version"]
                .asDict()
            )
        except:
            databricks_version: dict = {}

        return version, databricks_version

    def _get_deltalake_flow(
        self, app_name: str, name: str, config: dict, save_type: Type[Save]
    ):

        dataflow_config: dict = cp.load_pipeline_config(app_name, name)
        dataflow_config = dataflow_config.get("dataflow")

        self.log.debug("Deserializing configuration into Dataflow")

        dataflow = Dataflow(self, config, dataflow_config, save_type)

        return dataflow

    def _get_spark_context(self, app_name: str, config: dict):
        self.log.info("Setting spark context")
        spark_config = config["spark"]

        msg = json.dumps(spark_config, indent=4, default=str)
        self.log.debug(msg)

        builder = SparkSession.builder

        for k, v in spark_config.items():
            builder = builder.config(k, v)

        builder.appName(app_name)
        spark = configure_spark_with_delta_pip(builder).getOrCreate()

        # spark = builder.appName(app_name).getOrCreate()

        return spark

    def _get_spark_logger(self, spark: SparkSession, app_name: str, log_level: str):
        sc = spark.sparkContext
        sc.setLogLevel(log_level)
        log4j_logger = sc._jvm.org.apache.log4j
        logger = log4j_logger.LogManager.getLogger(app_name)

        return logger



from enum import Enum

BAD_RECORDS_PATH = "badRecordsPath"
FORMAT = "format"
MODE = "mode"
OPTIONS = "options"
PATH = "path"
EXCEPTIONS = "exceptions"
TABLE = "table"
DATABASE = "database"
PERMISSIVE = "permissive"
CORRUPT_RECORD = "_corrupt_record"
MERGE_SCHEMA = "mergeSchema"
APPEND = "append"
NAME = "name"
REPLACE = "replace"
ARGS = "args"
AUTO_IO = "auto_io"
TIMESTAMP = "timestamp"
INFER_SCHEMA = "inferSchema"
RECURSIVE_FILE_LOOKUP = "recursiveFileLookup"
CHECK_CONSTRAINTS = "check_constraints"
PROPERTIES = "properties"
PARTITIONS = "partitions"

# lineage columns
CONTEXT_ID = "_context_id"
LOAD_TIMESTAMP = "_load_timestamp"
TIMESLICE = "_timeslice"
FILENAME = "_filename"

# yetl table properties
YETL_TBLP_METADATA_LINEAGE_ENABLED = "yetl.metadata.lineageEnabled"
YETL_TBLP_SCHEMA_CREATE_IF_NOT_EXISTS = "yetl.schema.createIfNotExists"
YETL_TBLP_METADATA_TIMESLICE = "yetl.metadata.timeslice"


class Format(Enum):
    JSON = "json"
    CSV = "csv"
    PARQUET = "parquet"
    DELTA = "delta"



import json
from pyspark.sql import DataFrame
from enum import Enum


class DeltaLakeProperties(Enum):
    OPTIMIZE_WRITE = "delta.autoOptimize.optimizeWrite"
    AUTO_COMPACT = "delta.autoOptimize.autoCompact"


def get_partition_predicate(partition_values: dict):

    predicates = []
    for k, v in partition_values.items():
        v = [str(val) for val in v]
        p = f"`{k}` in ({','.join(v)})"
        predicates.append(p)

    predicate = " and ".join(predicates)

    return predicate


def table_exists(context, database: str, table: str):
    table_exists = (
        context.spark.sql(f"SHOW TABLES in {database};")
        .where(f"tableName='{table}' AND !isTemporary")
        .count()
        == 1
    )
    return table_exists


def create_table(context, database: str, table: str, path: str, sql: str = None):
    context.log.info(f"Creating table if not exists {database}.{table} at {path}")

    if not sql:
        sql = f"""
            CREATE TABLE IF NOT EXISTS `{database}`.`{table}` 
            USING DELTA LOCATION '{path}';"""

    context.spark.sql(sql)


def create_database(context, database: str):
    context.log.info(f"Creating database if not exists `{database}`")
    sql = f"CREATE DATABASE IF NOT EXISTS {database}"
    context.log.debug(sql)
    context.spark.sql(sql)


def get_audit(context, database_table: str):
    context.log.info(f"Auditing database table {database_table}")
    sql = f"DESCRIBE HISTORY {database_table}"
    context.log.debug(sql)
    audit = context.spark.sql(sql).first().asDict()
    return audit


def get_audits(context):
    audit = {}
    for database_table, destination in context.dataflow.destinations.items():
        audit[database_table] = get_audit(context, destination.database_table)

    results = json.dumps(audit, indent=4, sort_keys=True, default=str)

    context.log.info(results)
    return audit


def alter_table_drop_constraint(database: str, table: str, name: str):

    return f"ALTER TABLE `{database}`.`{table}` DROP CONSTRAINT {name};"


def alter_table_add_constraint(database: str, table: str, name: str, constraint: str):

    return f"ALTER TABLE `{database}`.`{table}` ADD CONSTRAINT {name} CHECK ({constraint});"


def alter_table_set_tblproperties(database: str, table: str, properties: str):

    return f"ALTER TABLE `{database}`.`{table}` SET TBLPROPERTIES ({properties});"


def get_table_properties(context, database: str, table: str):

    context.log.info(f"getting existing table properties for table {database}.{table}")
    df: DataFrame = context.spark.sql(
        f"SHOW TBLPROPERTIES `{database}`.`{table}`"
    ).collect()
    tbl_properties = {
        c.asDict()["key"]: c.asDict()["value"]
        for c in df
        if not c["key"].startswith("delta.constraints")
    }
    tbl_constraints = {
        c.asDict()["key"].split(".")[-1]: c.asDict()["value"]
        for c in df
        if c["key"].startswith("delta.constraints")
    }

    properties = {
        f"{database}.{table}": {
            "constraints": tbl_constraints,
            "properties": tbl_properties,
        }
    }
    msg = json.dumps(properties, indent=4, default=str)
    context.log.info(msg)
    return properties


def optimize(
    context, database: str, table: str, partition_values: dict, zorder_by: list = []
):
    sql = f"OPTIMIZE `{database}`.`{table}`"

    if partition_values:
        predicate = get_partition_predicate(partition_values)
        predicate = f" WHERE {predicate}"
        sql = f"{sql}{predicate}"

    if zorder_by:
        sql_zorderby = ",".join([f"`{z}`" for z in zorder_by])
        sql = f"{sql} ZORDER BY ({sql_zorderby})"

    context.log.info(f"optimizing table {database}.{table}\n{sql}")
    context.spark.sql(sql)


def get_table_details(context, database: str, table: str):

    context.log.info(
        f"getting existing table details and partitions for table {database}.{table}"
    )

    df: DataFrame = context.spark.sql(
        f"DESCRIBE TABLE EXTENDED `{database}`.`{table}`"
    ).collect()

    # get the details into a dictionary
    details = {c.asDict()["col_name"]: c.asDict()["data_type"] for c in df}

    # pull out the columns
    columns = {}
    ordinal = 0
    for k, v in details.items():
        if k and v:
            columns[k] = {}
            columns[k]["ordinal"] = ordinal
            columns[k]["type"] = v
            ordinal = +1
        else:
            break

    # pull out the columns
    partitions = [v for k, v in details.items() if k.startswith("Part ")]

    details = {
        f"{database}.{table}": {
            "columns": columns,
            "partitions": partitions,
            "name": details.get("Name"),
            "location": details.get("Location"),
            "provider": details.get("Provider"),
            "owner": details.get("Owner"),
        }
    }

    msg = json.dumps(details, indent=4, default=str)
    context.log.info(msg)
    return details



from pyspark.sql.types import StructType


class ISchemaRepo:
    def __init__(self, context, config: dict) -> None:
        self.context = context

    def save_schema(self, schema: StructType, database_name: str, table_name: str):
        """Save a schema into the repo."""
        pass

    def load_schema(self, database_name: str, table_name: str):
        """Loads a spark from a file."""
        pass



from ..dataset import Dataset
from datetime import datetime


class IMetadataRepo:
    def __init__(self, context, config: dict) -> None:
        self.context = context

    def _get_dataset(self, dataset: Dataset):
        metadata = {
            str(dataset.id): {
                "context_id": str(dataset.context_id),
                "database": dataset.database,
                "table": dataset.table,
                "load_timestamp": datetime.now().strftime("%Y-%m-%d %H:%M:%S"),
                "load_timestamp_utc": datetime.utcnow().strftime("%Y-%m-%d %H:%M:%S"),
                "path": dataset.path,
                "timeslice": "",
            }
        }

        return metadata

    def _get_index(self, dataset: Dataset):

        depends_on_ids = []
        # add the source dataset id's into the attribute list
        if dataset.is_destination():
            sources: dict = self.context.dataflow.sources
            depends_on_ids = [str(v.id) for k, v in sources.items()]

        index = {str(dataset.id): {"depends_on": depends_on_ids}}
        return index

    def save(self, dataset: Dataset):
        """Save a schema into the repo."""
        pass



from ..parser.parser import parse_functions
from ..parser._constants import NAME, REPLACE, ARGS


def execute_replacements(sentence: str, dataset):

    parsed: dict = parse_functions(sentence)

    for fn in parsed:

        function_name = fn.get(NAME)
        replace = fn.get(REPLACE)
        args = fn.get(ARGS)

        f = built_in_functions.get(function_name)
        result = f(function_name, dataset, *args)
        sentence = sentence.replace(replace, result)

    sentence = sentence.replace("{{", "").replace("}}", "")

    return sentence


def timeslice(config_name: str, dataset, attribute: str):

    formatted = None

    if hasattr(dataset, attribute) and hasattr(dataset.context, config_name):
        format_str = getattr(dataset, attribute)
        formatted = dataset.context.timeslice.strftime(format_str)

    return formatted


built_in_functions = {"timeslice": timeslice}



from .dataset import (
    Reader,
    Writer,
    StreamReader,
    StreamWriter,
    MergeSave,
    AppendSave,
    OverwriteSave,
    OverwriteSchemaSave,
    IgnoreSave,
    ErrorIfExistsSave,
    Save,
)
from .dataflow import IDataflow
from ._framework import yetl_flow
from ._context import Context
from ._timeslice import Timeslice, TimesliceUtcNow, TimesliceNow


__all__ = [
    "Reader",
    "Writer",
    "StreamReader",
    "StreamWriter",
    "yetl_flow",
    "Context",
    "IDataflow",
    "Timeslice",
    "TimesliceUtcNow",
    "TimesliceNow",
    "MergeSave",
    "AppendSave",
    "OverwriteSave",
    "OverwriteSchemaSave" "IgnoreSave",
    "ErrorIfExistsSave",
    "Save",
]



from . import _logging_config  # must be the 1st import
from ._context import Context
from ._delta_lake import get_audits
import time
from typing import Type
from .dataset import Save, DefaultSave


class YetlFlowException(Exception):
    def __init__(self, message):
        super().__init__(message)
        self.message = message


def yetl_flow(name: str = None, app_name: str = None, log_level="INFO"):
    def decorate(function):
        def wrap_function(*args, **kwargs):

            # default the name to the function name of the deltaflow
            if not name:
                _name = function.__name__

            spark = kwargs.get("spark")
            timeslice = kwargs.get("timeslice")
            save_type: Type[Save] = kwargs.get("save_type", DefaultSave)
            if "timeslice" in kwargs.keys():
                del kwargs["timeslice"]
            if "save_type" in kwargs.keys():
                del kwargs["save_type"]

            # create the context for the pipeline to run
            context = Context(app_name, log_level, _name, spark, timeslice, save_type)

            # run the pipeline
            yetl_flow_exception = None
            context.log.info(
                f"""Executing Dataflow {context.app_name} with:
                timeslice={timeslice} 
                retries={context.dataflow.retries} 
                save_type={save_type.__name__}"""
            )
            while context.dataflow.retries >= 0:
                try:
                    function(
                        context=context,
                        dataflow=context.dataflow,
                        timeslice=timeslice,
                        save_type=save_type,
                        *args,
                        **kwargs,
                    )
                    break
                except Exception as e:
                    msg = f"Dataflow application {context.app_name} failed due to {e}."
                    context.dataflow.retries -= 1
                    context.log.error(msg)
                    yetl_flow_exception = e
                    if context.dataflow.retries >= 0:
                        context.log.info(
                            f"Retrying {context.app_name} in after {context.dataflow.retry_wait} seconds; {context.dataflow.retries} retries are remaning."
                        )
                        time.sleep(context.dataflow.retry_wait)

            if yetl_flow_exception:
                msg = f"Dataflow application {context.app_name} failed due to {yetl_flow_exception}. No retries remaining."
                raise YetlFlowException(msg) from yetl_flow_exception

            # get the delta lake audit information and add it to the return
            audit = get_audits(context)
            audit = {"deltalake_log": audit}

            return audit

        return wrap_function

    return decorate
